import os

SCHEMA_DIR = os.path.dirname(os.path.abspath(__file__))

# ./schema.json
SCHEMA_FILE = os.path.join(SCHEMA_DIR, "schema.json")
# ./schema_new.json
SCHEMA_NEW_FILE = os.path.join(SCHEMA_DIR, "schema_new.json")
